@@ -1,4 +1,4 @@
-/*      $Id: frame.c 23403 2006-10-13 17:35:24Z olivier $
+/*      $Id: frame.c 25530 2007-04-11 22:39:59Z olivier $
 
         This program is free software; you can redistribute it and/or modify
         it under the terms of the GNU General Public License as published by
@@ -37,6 +37,10 @@
 #include "frame.h"
 #include "compositor.h"
 
+#ifndef ShapeInput
+#define ShapeInput 2
+#endif
+
 int
 frameDecorationLeft (ScreenInfo *screen_info)
 {
@@ -585,7 +589,7 @@ frameSetShape (Client * c, int state, FramePixmap * frame_pix, int button_x[BUTT
 {
     ScreenInfo *screen_info;
     DisplayInfo *display_info;
-    Window temp;
+    Window shape_win;
     XRectangle rect;
     xfwmPixmap *my_pixmap;
     int i;
@@ -601,7 +605,7 @@ frameSetShape (Client * c, int state, FramePixmap * frame_pix, int button_x[BUTT
         return;
     }
 
-    temp = XCreateSimpleWindow (display_info->dpy, c->frame, 0, 0, frameWidth (c), frameHeight (c), 0, 0, 0);
+    shape_win = XCreateSimpleWindow (display_info->dpy, c->frame, 0, 0, frameWidth (c), frameHeight (c), 0, 0, 0);
 
     if (FLAG_TEST (c->flags, CLIENT_FLAG_SHADED))
     {
@@ -609,12 +613,19 @@ frameSetShape (Client * c, int state, FramePixmap * frame_pix, int button_x[BUTT
         rect.y = 0;
         rect.width  = frameWidth (c);
         rect.height = frameHeight (c);
-        XShapeCombineRectangles (display_info->dpy, temp, ShapeBounding, 0, 0, &rect, 1,
-                                 ShapeSubtract, 0);
+        XShapeCombineRectangles (display_info->dpy, shape_win, ShapeBounding, 0, 0, &rect, 1, ShapeSubtract, Unsorted);
+    }
+    else if (!FLAG_TEST (c->flags, CLIENT_FLAG_HAS_SHAPE))
+    {
+        rect.x = frameLeft (c);
+        rect.y = frameTop (c);
+        rect.width  = c->width;
+        rect.height = c->height;
+        XShapeCombineRectangles (display_info->dpy, shape_win, ShapeBounding, 0, 0, &rect, 1, ShapeSet, Unsorted);
     }
     else
     {
-        XShapeCombineShape (display_info->dpy, temp, ShapeBounding, frameLeft (c),
+        XShapeCombineShape (display_info->dpy, shape_win, ShapeBounding, frameLeft (c),
                             frameTop (c), c->window, ShapeBounding, ShapeSet);
     }
     if (frame_pix)
@@ -728,20 +739,20 @@ frameSetShape (Client * c, int state, FramePixmap * frame_pix, int button_x[BUTT
         {
             if (xfwmWindowVisible (&c->sides[SIDE_LEFT]))
             {
-                XShapeCombineShape (display_info->dpy, temp, ShapeBounding, 0, frameTop (c),
+                XShapeCombineShape (display_info->dpy, shape_win, ShapeBounding, 0, frameTop (c),
                                     MYWINDOW_XWINDOW (c->sides[SIDE_LEFT]), ShapeBounding, ShapeUnion);
             }
 
             if (xfwmWindowVisible (&c->sides[SIDE_RIGHT]))
             {
-                XShapeCombineShape (display_info->dpy, temp, ShapeBounding, frameWidth (c) - frameRight (c), frameTop (c),
+                XShapeCombineShape (display_info->dpy, shape_win, ShapeBounding, frameWidth (c) - frameRight (c), frameTop (c),
                                     MYWINDOW_XWINDOW (c->sides[SIDE_RIGHT]), ShapeBounding, ShapeUnion);
             }
         }
 
         if (xfwmWindowVisible (&c->title))
         {
-            XShapeCombineShape (display_info->dpy, temp, ShapeBounding,
+            XShapeCombineShape (display_info->dpy, shape_win, ShapeBounding,
                                 frameTopLeftWidth (c, state), 0,
                                 MYWINDOW_XWINDOW (c->title), ShapeBounding, ShapeUnion);
         }
@@ -749,13 +760,13 @@ frameSetShape (Client * c, int state, FramePixmap * frame_pix, int button_x[BUTT
         if (xfwmWindowVisible (&c->corners[CORNER_TOP_LEFT]))
         {
 
-            XShapeCombineShape (display_info->dpy, temp, ShapeBounding, 0, 0,
+            XShapeCombineShape (display_info->dpy, shape_win, ShapeBounding, 0, 0,
                                 MYWINDOW_XWINDOW (c->corners[CORNER_TOP_LEFT]), ShapeBounding, ShapeUnion);
         }
 
         if (xfwmWindowVisible (&c->sides[SIDE_BOTTOM]))
         {
-            XShapeCombineShape (display_info->dpy, temp, ShapeBounding,
+            XShapeCombineShape (display_info->dpy, shape_win, ShapeBounding,
                                 screen_info->corners[CORNER_BOTTOM_LEFT][state].width,
                                 frameHeight (c) - frameBottom (c),
                                 MYWINDOW_XWINDOW (c->sides[SIDE_BOTTOM]), ShapeBounding, ShapeUnion);
@@ -763,14 +774,14 @@ frameSetShape (Client * c, int state, FramePixmap * frame_pix, int button_x[BUTT
 
         if (xfwmWindowVisible (&c->corners[CORNER_BOTTOM_LEFT]))
         {
-            XShapeCombineShape (display_info->dpy, temp, ShapeBounding, 0,
+            XShapeCombineShape (display_info->dpy, shape_win, ShapeBounding, 0,
                                 frameHeight (c) - screen_info->corners[CORNER_BOTTOM_LEFT][state].height,
                                 MYWINDOW_XWINDOW (c->corners[CORNER_BOTTOM_LEFT]), ShapeBounding, ShapeUnion);
         }
 
         if (xfwmWindowVisible (&c->corners[CORNER_BOTTOM_RIGHT]))
         {
-            XShapeCombineShape (display_info->dpy, temp, ShapeBounding,
+            XShapeCombineShape (display_info->dpy, shape_win, ShapeBounding,
                                 frameWidth (c) - screen_info->corners[CORNER_BOTTOM_RIGHT][state].width,
                                 frameHeight (c) - screen_info->corners[CORNER_BOTTOM_RIGHT][state].height,
                                 MYWINDOW_XWINDOW (c->corners[CORNER_BOTTOM_RIGHT]), ShapeBounding, ShapeUnion);
@@ -778,7 +789,7 @@ frameSetShape (Client * c, int state, FramePixmap * frame_pix, int button_x[BUTT
 
         if (xfwmWindowVisible (&c->corners[CORNER_TOP_RIGHT]))
         {
-            XShapeCombineShape (display_info->dpy, temp, ShapeBounding,
+            XShapeCombineShape (display_info->dpy, shape_win, ShapeBounding,
                                 frameWidth (c) - frameTopRightWidth (c, state),
                                 0, MYWINDOW_XWINDOW (c->corners[CORNER_TOP_RIGHT]), ShapeBounding, ShapeUnion);
         }
@@ -787,21 +798,27 @@ frameSetShape (Client * c, int state, FramePixmap * frame_pix, int button_x[BUTT
         {
             if (xfwmWindowVisible (&c->buttons[i]))
             {
-                XShapeCombineShape (display_info->dpy, temp, ShapeBounding, button_x[i],
+                XShapeCombineShape (display_info->dpy, shape_win, ShapeBounding, button_x[i],
                                     (frameTop (c) - screen_info->buttons[i][state].height + 1) / 2,
                                     MYWINDOW_XWINDOW (c->buttons[i]), ShapeBounding, ShapeUnion);
             }
         }
     }
-
     rect.x = 0;
     rect.y = 0;
     rect.width  = frameWidth (c);
     rect.height = frameHeight (c);
-    XShapeCombineRectangles (display_info->dpy, temp, ShapeBounding, 0, 0, &rect, 1, ShapeIntersect, 0);
-    XShapeCombineShape (display_info->dpy, c->frame, ShapeBounding, 0, 0, temp, ShapeBounding, ShapeSet);
+    XShapeCombineRectangles (display_info->dpy, shape_win, ShapeBounding, 0, 0, &rect, 1, ShapeIntersect, Unsorted);
+    XShapeCombineShape (display_info->dpy, c->frame, ShapeBounding, 0, 0, shape_win, ShapeBounding, ShapeSet);
 
-    XDestroyWindow (display_info->dpy, temp);
+    /* Set Input shape when using XShape extension 1.1 and later */
+    if (display_info->shape_version >= 1001)
+    {
+        XShapeCombineShape (display_info->dpy, shape_win, ShapeInput, frameLeft (c), frameTop (c), c->window, ShapeBounding, ShapeSubtract);
+        XShapeCombineShape (display_info->dpy, shape_win, ShapeInput, frameLeft (c), frameTop (c), c->window, ShapeInput, ShapeUnion);
+        XShapeCombineShape (display_info->dpy, c->frame, ShapeInput, 0, 0, shape_win, ShapeInput, ShapeSet);
+    }
+    XDestroyWindow (display_info->dpy, shape_win);
 }
 
 void
@@ -870,6 +887,13 @@ frameDraw (Client * c, gboolean clear_all)
         width_changed = TRUE;
         height_changed = TRUE;
         requires_clearing = TRUE;
+        frameClearQueueDraw (c);
+        if (c->frame_timeout_id)
+        {
+            g_source_remove (c->frame_timeout_id);
+            c->frame_timeout_id = 0;
+        }
+
     }
     else
     {
@@ -1124,3 +1148,48 @@ frameDraw (Client * c, gboolean clear_all)
         frameSetShape (c, 0, NULL, 0);
     }
 }
+
+static gboolean
+update_frame_idle_cb (gpointer data)
+{
+    Client *c;
+    
+    TRACE ("entering update_frame_idle_cb");
+
+    c = (Client *) data;
+    g_return_val_if_fail (c, FALSE);
+
+    frameDraw (c, TRUE);
+    c->frame_timeout_id = 0;
+
+    return FALSE;
+}
+
+void
+frameClearQueueDraw (Client * c)
+{
+    g_return_if_fail (c);
+
+    TRACE ("entering frameClearQueueDraw for \"%s\" (0x%lx)", c->name, c->window);
+
+    if (c->frame_timeout_id)
+    {
+        g_source_remove (c->frame_timeout_id);
+        c->frame_timeout_id = 0;
+    }
+}
+
+void
+frameQueueDraw (Client * c)
+{
+    g_return_if_fail (c);
+
+    TRACE ("entering frameQueueDraw for \"%s\" (0x%lx)", c->name, c->window);
+
+    if (c->frame_timeout_id == 0)
+    {
+        c->frame_timeout_id = g_idle_add_full (G_PRIORITY_DEFAULT_IDLE, 
+                                              update_frame_idle_cb, c, NULL);
+    }
+}
+
